/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.springframework.cloud.alibaba.dubbo.autoconfigure;

import org.apache.dubbo.common.URL;
import org.apache.dubbo.config.ApplicationConfig;
import org.apache.dubbo.config.ProtocolConfig;
import org.apache.dubbo.config.ServiceConfig;
import org.apache.dubbo.config.spring.ServiceBean;
import org.apache.dubbo.config.spring.context.event.ServiceBeanExportedEvent;

import com.ecwid.consul.v1.agent.model.NewService;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.autoconfigure.AutoConfigureAfter;
import org.springframework.boot.autoconfigure.condition.ConditionalOnClass;
import org.springframework.boot.context.event.ApplicationFailedEvent;
import org.springframework.cloud.alibaba.dubbo.metadata.repository.DubboServiceMetadataRepository;
import org.springframework.cloud.alibaba.dubbo.metadata.resolver.MetadataResolver;
import org.springframework.cloud.alibaba.dubbo.registry.event.ServiceInstancePreRegisteredEvent;
import org.springframework.cloud.alibaba.dubbo.service.DubboMetadataConfigService;
import org.springframework.cloud.alibaba.dubbo.service.PublishingDubboMetadataConfigService;
import org.springframework.cloud.alibaba.dubbo.util.JSONUtils;
import org.springframework.cloud.client.ServiceInstance;
import org.springframework.cloud.client.serviceregistry.Registration;
import org.springframework.cloud.client.serviceregistry.ServiceRegistry;
import org.springframework.cloud.consul.serviceregistry.ConsulRegistration;
import org.springframework.context.ApplicationListener;
import org.springframework.context.ConfigurableApplicationContext;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.event.ContextClosedEvent;
import org.springframework.context.event.EventListener;
import org.springframework.util.CollectionUtils;
import org.springframework.util.StringUtils;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.function.Supplier;
import java.util.stream.Collectors;

import static org.springframework.cloud.alibaba.dubbo.registry.SpringCloudRegistry.DUBBO_URLS_METADATA_PROPERTY_NAME;

/**
 * The Auto-Configuration class for Dubbo metadata {@link EventListener event handling}.
 *
 * @author <a href="mailto:mercyblitz@gmail.com">Mercy</a>
 */
@AutoConfigureAfter(value = {DubboMetadataAutoConfiguration.class})
@Configuration
public class DubboMetadataEventHandlingAutoConfiguration {

    private static final String CONSUL_REGISTRATION_CLASS_NAME =
            "org.springframework.cloud.consul.serviceregistry.ConsulAutoRegistration";

    private static final String EUREKA_REGISTRATION_CLASS_NAME =
            "org.springframework.cloud.netflix.eureka.serviceregistry.EurekaAutoServiceRegistration";

    private final Logger logger = LoggerFactory.getLogger(getClass());

    @Autowired
    private MetadataResolver metadataResolver;

    @Autowired
    private PublishingDubboMetadataConfigService dubboMetadataConfigService;

    @Autowired
    private ApplicationConfig applicationConfig;

    @Autowired
    private Supplier<ProtocolConfig> protocolConfigSupplier;

    @Autowired
    private ConfigurableApplicationContext context;

    @Autowired
    private DubboServiceMetadataRepository dubboServiceMetadataRepository;

    @Autowired
    private JSONUtils jsonUtils;

    @Value("${spring.application.name:application}")
    private String currentApplicationName;

    @Autowired
    private ServiceRegistry serviceRegistry;

    private volatile Registration registration;

    /**
     * The ServiceConfig of DubboMetadataConfigService to be exported, can be nullable.
     */
    private ServiceConfig<DubboMetadataConfigService> serviceConfig;

    @EventListener(ServiceBeanExportedEvent.class)
    public void onServiceBeanExported(ServiceBeanExportedEvent event) {
        ServiceBean serviceBean = event.getServiceBean();
        publishServiceRestMetadata(serviceBean);
    }

    @ConditionalOnClass(name = EUREKA_REGISTRATION_CLASS_NAME)
    @Bean
    public ApplicationListener<ServiceBeanExportedEvent> onServiceBeanExportedInEureka() {
        return event -> {
            reRegister();
        };
    }

    private void reRegister() {
        Registration registration = this.registration;
        if (registration == null) {
            return;
        }
        serviceRegistry.register(registration);
    }

    @EventListener(ApplicationFailedEvent.class)
    public void onApplicationFailed() {
        unexportDubboMetadataConfigService();
    }

    @EventListener(ContextClosedEvent.class)
    public void onContextClosed() {
        unexportDubboMetadataConfigService();
    }

    @EventListener(ServiceInstancePreRegisteredEvent.class)
    public void onServiceInstancePreRegistered(ServiceInstancePreRegisteredEvent event) {
        Registration registration = event.getSource();
        exportDubboMetadataConfigService();
        attachURLsIntoMetadata(registration);
        setRegistration(registration);
    }

    private void setRegistration(Registration registration) {
        this.registration = registration;
    }

    /**
     * Handle the pre-registered event of {@link ServiceInstance} for Consul
     *
     * @return ApplicationListener<ServiceInstancePreRegisteredEvent>
     */
    @ConditionalOnClass(name = CONSUL_REGISTRATION_CLASS_NAME)
    @Bean
    public ApplicationListener<ServiceInstancePreRegisteredEvent> onServiceInstancePreRegisteredInConsul() {
        return event -> {
            Registration registration = event.getSource();
            String registrationClassName = registration.getClass().getName();
            if (CONSUL_REGISTRATION_CLASS_NAME.equalsIgnoreCase(registrationClassName)) {
                NewService newService = ((ConsulRegistration) registration).getService();
                String dubboURLsJson = getDubboURLsJSON();
                if (StringUtils.hasText(dubboURLsJson)) {
                    List<String> tags = newService.getTags();
                    tags.add(DUBBO_URLS_METADATA_PROPERTY_NAME + "=" + dubboURLsJson);
                }
            }
        };
    }

    private void attachURLsIntoMetadata(Registration registration) {
        if (registration == null) {
            return;
        }
        Map<String, String> metadata = registration.getMetadata();
        String dubboURLsJson = getDubboURLsJSON();
        if (StringUtils.hasText(dubboURLsJson)) {
            metadata.put(DUBBO_URLS_METADATA_PROPERTY_NAME, dubboURLsJson);
        }
    }

    private String getDubboURLsJSON() {
        Collection<URL> urls = dubboServiceMetadataRepository.getRegisteredUrls();
        if (CollectionUtils.isEmpty(urls)) {
            if (logger.isDebugEnabled()) {
                logger.debug("There is no registered URL to attach into metadata.");
            }
            return null;
        }
        return jsonUtils.toJSON(urls.stream().map(URL::toFullString).collect(Collectors.toList()));
    }

    private void publishServiceRestMetadata(ServiceBean serviceBean) {
        dubboMetadataConfigService.publishServiceRestMetadata(metadataResolver.resolveServiceRestMetadata(serviceBean));
    }

    private void exportDubboMetadataConfigService() {

        if (serviceConfig != null && serviceConfig.isExported()) {
            return;
        }

        if (StringUtils.isEmpty(dubboMetadataConfigService.getServiceRestMetadata())) {
            // If there is no REST metadata, DubboMetadataConfigService will not be exported.
            if (logger.isInfoEnabled()) {
                logger.info("There is no REST metadata, the Dubbo service[{}] will not be exported.",
                        dubboMetadataConfigService.getClass().getName());
            }
            return;
        }

        serviceConfig = new ServiceConfig<>();

        serviceConfig.setInterface(DubboMetadataConfigService.class);
        // Use current Spring application name as the Dubbo Service version
        serviceConfig.setVersion(currentApplicationName);
        serviceConfig.setRef(dubboMetadataConfigService);
        serviceConfig.setApplication(applicationConfig);
        serviceConfig.setProtocol(protocolConfigSupplier.get());

        serviceConfig.export();

        if (logger.isInfoEnabled()) {
            logger.info("The Dubbo service[{}] has been exported.", serviceConfig.toString());
        }
    }

    private void unexportDubboMetadataConfigService() {

        if (serviceConfig == null || serviceConfig.isUnexported()) {
            return;
        }

        serviceConfig.unexport();

        if (logger.isInfoEnabled()) {
            logger.info("The Dubbo service[{}] has been unexported.", serviceConfig.toString());
        }
    }
}
